import numpy as np
import math
from numba import njit, jit

# from scipy.stats import chi2
# from lifelines import CoxPHFitter
# from lifelines.statistics import logrank_test
# import fastlogranktest as flr

"""
SIGNIFICANCE EVALUATION (chi2.sf, chi2.isf)
MODIFIED CODE FROM 
1. https://github.com/etal/biofrills/blob/36684bb6c7632f96215e8b2b4ebc86640f331bcd/biofrills/stats/chisq.py
2. https://github.com/scipy/scipy/blob/v1.13.0/scipy/stats/_continuous_distns.py#L1547
3. https://github.com/scipy/scipy/blob/249c11f12ca11b000490da615c3008b2dd213322/scipy/special/special/cephes/igam.h
4. https://github.com/catboost/catboost/blob/f2c5eb6a1b1e1e936b56b3e523bd8dfd77b46fb9/contrib/python/scipy/py3/scipy/special/cephes/lanczos.h#L69
"""

MACHEP = 1e-13  # the machine roundoff error / tolerance
BORDER = 4e15
BORDERINV = 1 / BORDER

EULER = 0.577215664
MAXLOG = 709.79
LANCZOS_G = 6.02468

MAXITER = 2000
LGAMMA_05 = 0.5723649429247004


@njit
def _igamc(a, x):
    """
    In this implementation both arguments must be positive.
    The integral is evaluated by either a power series or
    continued fraction expansion, depending on the relative
    values of a and x.
    """
    # Compute  x**a * exp(-x) / Gamma(a)
    ax = math.exp(a * math.log(x) - x - LGAMMA_05)

    # Continued fraction
    y = 1.0 - a
    z = x + y + 1.0
    c = 0.0
    pkm2 = 1.0
    qkm2 = x
    pkm1 = x + 1.0
    qkm1 = z * x
    ans = pkm1 / qkm1
    while True:
        c += 1.0
        y += 1.0
        z += 2.0
        yc = y * c
        pk = pkm1 * z - pkm2 * yc
        qk = qkm1 * z - qkm2 * yc
        if qk != 0:
            r = pk / qk
            t = abs((ans - r) / r)
            ans = r
        else:
            t = 1.0
        pkm2 = pkm1
        pkm1 = pk
        qkm2 = qkm1
        qkm1 = qk
        if abs(pk) > BORDER:
            pkm2 *= BORDERINV
            pkm1 *= BORDERINV
            qkm2 *= BORDERINV
            qkm1 *= BORDERINV
        if t <= MACHEP:
            return ans * ax


@njit
def power_series(a, x):
    r = a
    c = 1.0
    ans = 1.0

    for i in range(0, MAXITER):
        r += 1.0
        c *= x / r
        ans += c
        if c <= MACHEP * ans:
            break
    return ans / a


@njit
def _igam(a, x):
    """ Left tail of incomplete Gamma function """
    # Compute  x**a * exp(-x) / Gamma(a)
    ax = math.exp(a * math.log(x) - x - LGAMMA_05)
    return power_series(a, x) * ax


@njit
def igam_series(a, x):
    ax = igam_fac(a, x)
    if ax == 0.0:
        return 0.0
    return power_series(a, x) * ax


@njit
def chi2_sf(x, df):
    """
    Probability value (1-tail) for the Chi^2 probability distribution.

    Broadcasting rules apply.

    Parameters
    ----------
    x : array_like or float > 0
    df : array_like or float, probably int >= 1

    Returns
    -------
    chisqprob : ndarray
        The area from `chisq` to infinity under the Chi^2 probability
        distribution with degrees of freedom `df`.

    """
    if x <= 0:
        return 1.0
    if x == 0:
        return 0.0
    if df <= 0:
        raise ValueError("Domain error.")
    if x < 1.0 or x < df:
        return 1.0 - _igam(0.5 * df, 0.5 * x)
    return _igamc(0.5 * df, 0.5 * x)


@njit('f4(f4[:], f4)', cache=True)
def polyval(p, x):
    """
    Evaluate a polynomial by Horner's scheme
    """
    y = 0
    for pv in p:
        y = y * x + pv
    return y


@njit
def ratevl(x, num, denom):  # N = M = 12
    absx = np.abs(x)
    if absx > 1:
        '''Evaluate as a polynomial in 1/x.'''
        num_ans = polyval(num[::-1], 1 / x)
        denom_ans = polyval(denom[::-1], 1 / x)
        return np.power(x, 0) * num_ans / denom_ans
    else:
        num_ans = polyval(num, x)
        denom_ans = polyval(denom, x)
    return num_ans / denom_ans


@njit
def lanczos_sum_expg_scaled(x):
    lanczos_sum_expg_scaled_num = np.array([
        0.00606184234, 0.50984166556, 19.5199278824, 449.944556906,
        6955.99960251, 75999.2930401, 601859.617168, 3481712.15498, 14605578.0876,
        43338889.3246, 86363131.2881, 103794043.116, 56906521.9134], dtype=np.float32)

    lanczos_sum_expg_scaled_denom = np.array([
        1, 66, 1925, 32670, 357423, 2637558, 13339535, 45995730,
        105258076, 150917976, 120543840, 39916800, 0], dtype=np.float32)
    return ratevl(x, lanczos_sum_expg_scaled_num, lanczos_sum_expg_scaled_denom)


@njit
def igam_fac(a, x):
    if np.abs(a - x) > 0.4 * np.abs(a):
        ax = a * np.log(x) - x - LGAMMA_05
        if (ax < -MAXLOG):
            return 0.0
        return np.exp(ax)

    fac = a + LANCZOS_G - 0.5
    res = np.sqrt(fac / np.exp(1)) / lanczos_sum_expg_scaled(a)

    if (a < 200) and (x < 200):
        res *= np.exp(a - x) * pow(x / fac, a)
    else:
        num = x - a - LANCZOS_G + 0.5
        res *= np.exp(a * (np.log1p(num / fac) - num / fac) + x * (0.5 - LANCZOS_G) / fac)
    return res


@njit
def find_inverse_gamma(a, p, q):
    """
    In order to understand what's going on here, you will
    need to refer to:

    Computation of the Incomplete Gamma Function Ratios and their Inverse
    ARMIDO R. DIDONATO and ALFRED H. MORRIS, JR.
    ACM Transactions on Mathematical Software, Vol. 12, No. 4,
    December 1986, Pages 377-393.
    """
    if a == 1:
        if q > 0.9:
            result = -np.log1p(-p)
        else:
            result = -np.log(q)
    elif a < 1:
        g = math.gamma(a)
        b = q * g

        if ((b > 0.6) or ((b >= 0.45) and (a >= 0.3))):
            '''
            DiDonato & Morris Eq 21:

            There is a slight variation from DiDonato and Morris here:
            the first form given here is unstable when p is close to 1,
            making it impossible to compute the inverse of Q(a,x) for small
            q. Fortunately the second form works perfectly well in this case.
            '''
            if (b * q > 1e-8) and (q > 1e-5):
                u = np.power(p * g * a, 1 / a)
            else:
                u = np.exp((-q / a) - EULER)
            result = u / (1 - (u / (a + 1)))
        elif (a < 0.3) and (b >= 0.35):
            '''DiDonato & Morris Eq 22:'''
            t = np.exp(-EULER - b)
            u = t * np.exp(t)
            result = t * np.exp(u)
        elif (b > 0.15) or (a >= 0.3):
            ''' DiDonato & Morris Eq 23: '''
            y = -np.log(b)
            u = y - (1 - a) * np.log(y)
            result = y - (1 - a) * np.log(u) - np.log(1 + (1 - a) / (1 + u))
        elif b > 0.1:
            ''' DiDonato & Morris Eq 24: '''
            y = -np.log(b)
            u = y - (1 - a) * np.log(y)
            result = y - (1 - a) * np.log(u) - np.log(
                (u * u + 2 * (3 - a) * u + (2 - a) * (3 - a)) / (u * u + (5 - a) * u + 2))
        else:
            ''' DiDonato & Morris Eq 25: '''
            y = -np.log(b)
            c1 = (a - 1) * np.log(y)
            c1_2 = c1 * c1
            c1_3 = c1_2 * c1
            c1_4 = c1_2 * c1_2
            a_2 = a * a
            a_3 = a_2 * a

            c2 = (a - 1) * (1 + c1)
            c3 = (a - 1) * (-(c1_2 / 2) + (a - 2) * c1 + (3 * a - 5) / 2)
            c4 = (a - 1) * ((c1_3 / 3) - (3 * a - 5) * c1_2 / 2 + (a_2 - 6 * a + 7) * c1 +
                            (11 * a_2 - 46 * a + 47) / 6)
            c5 = (a - 1) * (-(c1_4 / 4) + (11 * a - 17) * c1_3 / 6 + (-3 * a_2 + 13 * a - 13) * c1_2 +
                            (2 * a_3 - 25 * a_2 + 72 * a - 61) * c1 / 2 +
                            (25 * a_3 - 195 * a_2 + 477 * a - 379) / 12)

            y_2 = y * y
            y_3 = y_2 * y
            y_4 = y_2 * y_2
            result = y + c1 + (c2 / y) + (c3 / y_2) + (c4 / y_3) + (c5 / y_4)
    else:
        result = np.nan  # not implemented for df > 2
    return result


@njit
def igam(a, x):
    if a == 0:
        return 1 if (x > 0) else np.nan
    elif x == 0:
        '''Zero integration limit'''
        return 0
    elif np.isinf(a):
        return np.nan if np.isinf(x) else 0
    elif np.isinf(x):
        return 1
    if (x > 1.0) and (x > a):
        return 1.0 - _igamc(a, x)
    return igam_series(a, x)


@njit
def igami(a, p):
    if np.isnan(a) or np.isnan(p):
        return np.nan
    elif p == 0.0:
        return 0.0
    elif p == 1.0:
        return np.inf
    elif p > 0.9:
        return igamci(a, 1 - p)

    x = find_inverse_gamma(a, p, 1 - p)
    '''Halley's method'''
    for i in range(0, 3):
        fac = igam_fac(a, x)
        if fac == 0.0:
            return x
        f_fp = (igam(a, x) - p) * x / fac
        '''The ratio of the first and second derivatives simplifies'''
        fpp_fp = -1.0 + (a - 1) / x
        if np.isinf(fpp_fp):
            '''Resort to Newton's method in the case of overflow'''
            x = x - f_fp
        else:
            x = x - f_fp / (1.0 - 0.5 * f_fp * fpp_fp)
    return x


@njit
def igamci(a, q):
    if q == 0.0:
        return np.inf
    elif q == 1.0:
        return 0.0
    elif q > 0.9:
        return igami(a, 1 - q)

    x = find_inverse_gamma(a, 1 - q, q)
    for i in range(0, 3):
        fac = igam_fac(a, x)
        if fac == 0.0:
            return x
        f_fp = (_igamc(a, x) - q) * x / (-fac)
        fpp_fp = -1.0 + (a - 1) / x
        if np.isinf(fpp_fp):
            x = x - f_fp
        else:
            x = x - f_fp / (1.0 - 0.5 * f_fp * fpp_fp)
    return x


@njit
def chi2_isf(y, df):
    x = igamci(0.5 * df, y)
    return 2.0 * x


""" LOG-RANK CRITERIA """


# Interface functions
def logrank(durations_A, durations_B, event_observed_A=None, event_observed_B=None):
    return weight_lr_fast(durations_A, durations_B, event_observed_A, event_observed_B)


def wilcoxon(durations_A, durations_B, event_observed_A=None, event_observed_B=None):
    return weight_lr_fast(durations_A, durations_B, event_observed_A, event_observed_B, weightings="wilcoxon")


def peto(durations_A, durations_B, event_observed_A=None, event_observed_B=None):
    return weight_lr_fast(durations_A, durations_B, event_observed_A, event_observed_B, weightings="peto")


def tarone_ware(durations_A, durations_B, event_observed_A=None, event_observed_B=None):
    return weight_lr_fast(durations_A, durations_B, event_observed_A, event_observed_B, weightings="tarone-ware")


CRITERIA_DICT = {
    "logrank": logrank,
    "wilcoxon": wilcoxon,
    "peto": peto,
    "tarone-ware": tarone_ware
}
""" dict: Available criteria in library and its realization """


def logrank_self(durations_A, durations_B, event_observed_A=None, event_observed_B=None) -> float:
    # 27.1 ms ± 1.26 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)
    try:
        event_times_A, event_times_B = (np.array(durations_A), np.array(durations_B))
        time = np.r_[event_times_A, event_times_B]
        group_indicator = np.r_[np.zeros(event_times_A.shape[0], dtype=int), np.ones(event_times_B.shape[0], dtype=int)]
        event = np.r_[event_observed_A, event_observed_B]
        n_samples = time.shape[0]
        groups, group_counts = np.unique(group_indicator, return_counts=True)
        n_groups = groups.shape[0]
        if n_groups == 1:
            raise ValueError("At least two groups must be specified, "
                             "but only one was provided.")

        # sort descending
        o = np.argsort(-time, kind="mergesort")
        x = group_indicator[o]
        event = event[o]
        time = time[o]

        at_risk = np.zeros(n_groups, dtype=np.int_)
        observed = np.zeros(n_groups, dtype=np.int_)
        expected = np.zeros(n_groups, dtype=np.float_)
        covar = np.zeros((n_groups, n_groups), dtype=np.float_)
        k = 0
        while k < n_samples:
            ti = time[k]
            total_events = 0
            while k < n_samples and ti == time[k]:
                idx = np.searchsorted(groups, x[k])
                if event[k]:
                    observed[idx] += 1
                    total_events += 1
                at_risk[idx] += 1
                k += 1

            if total_events != 0:
                total_at_risk = k
                expected += at_risk * (total_events / total_at_risk)
                if total_at_risk > 1:
                    multiplier = total_events * (total_at_risk - total_events) / (total_at_risk * (total_at_risk - 1))
                    for g1 in range(n_groups):
                        temp = at_risk[g1] * multiplier
                        covar[g1, g1] += temp
                        for g2 in range(n_groups):
                            covar[g1, g2] -= temp * at_risk[g2] / total_at_risk

        df = n_groups - 1
        zz = observed[:df] - expected[:df]
        chisq = np.linalg.solve(covar[:df, :df], zz).dot(zz)
        pval = chi2_sf(chisq, df)
    except:
        pval = 1.0
    return pval
    

# def cox(durations_A, durations_B, event_observed_A=None, event_observed_B=None) -> float:
#     dfA = pd.DataFrame({'E': event_observed_A, 'T': durations_A, 'group': 0})
#     dfB = pd.DataFrame({'E': event_observed_B, 'T': durations_B, 'group': 1})
#     df_ = pd.concat([dfA, dfB])
#
#     cph = CoxPHFitter().fit(df_, 'T', 'E')
#     return cph.log_likelihood_ratio_test().p_value


# 06/02/2022 Numpy+Numba
# accelerate 2.8 according to library criteria
@njit
def iterate_coeffs_t_j(dur_A, dur_B, cens_A, cens_B, t_j, weightings):
    N_1_j = (dur_A >= t_j).sum()
    N_2_j = (dur_B >= t_j).sum()
    if N_1_j == 0 or N_2_j == 0:
        return 0, 0, 0
    O_1_j = ((dur_A == t_j) * cens_A).sum()  # np.where(dur_A == t_j, cens_A,0).sum()
    O_2_j = ((dur_B == t_j) * cens_B).sum()  # np.where(dur_B == t_j, cens_B,0).sum()
    
    N_j = N_1_j + N_2_j
    O_j = O_1_j + O_2_j
    E_1_j = N_1_j*O_j/N_j
    w_j = 1
    if weightings == "wilcoxon":
        w_j = N_j
    elif weightings == "tarone-ware":
        w_j = np.sqrt(N_j)
    elif weightings == "peto":
        w_j = (1.0 - float(O_j)/(N_j+1))
    
    num = O_1_j - E_1_j
    denom = E_1_j*(N_j - O_j) * N_2_j/(N_j*(N_j - 1))
    return w_j, num, denom


@njit
def iterate_lr_statistic(dur_A, dur_B, cens_A, cens_B, times, weightings) -> float:
    res = np.zeros((times.shape[0], 3), dtype=np.float32)
    for j, t_j in enumerate(times):
        res[j] = iterate_coeffs_t_j(dur_A, dur_B, cens_A, cens_B, t_j, weightings)

    if weightings == "peto":
        res[:, 0] = np.cumprod(res[:, 0])
    # logrank = np.dot(res[:, 0], res[:, 1])**2 / np.dot(res[:, 0]*res[:, 0], res[:, 2])
    logrank = np.power((res[:, 0]*res[:, 1]).sum(), 2) / ((res[:, 0]*res[:, 0]*res[:, 2]).sum())
    return logrank


def iterate_weight_lr_fast(dur_A, dur_B, cens_A=None, cens_B=None, weightings="") -> float:
    try:
        if cens_A is None:
            cens_A = np.ones(dur_A.shape[0])
        if cens_B is None:
            cens_B = np.ones(dur_B.shape[0])

        #     a1 = np.unique(dur_A)
        #     a2 = np.unique(dur_B)
        #     times = np.unique(np.clip(np.union1d(a1,a2), 0, np.min([a1.max(), a2.max()])))
        times = np.union1d(np.unique(dur_A), np.unique(dur_B))
        logrank = lr_statistic(dur_A, dur_B, cens_A, cens_B, times, weightings)
        pvalue = chi2_sf(logrank, df=1)
        return pvalue
    except:
        return 1.0


# 11/08/2022 Numpy+Numba
# dur_A = np.random.choice(10000, 10000)
# cens_A = np.random.choice(2, 10000)
# dur_B = np.random.choice(10000, 10000)
# cens_B = np.random.choice(2, 10000)
# new version (for 10000 elem): 2.16 ms ± 15.2 µs per loop
# old version (for 10000 elem): 103 ms ± 171 µs per loop


# @cuda.jit
@njit('f8(f8[:], f8[:], i8[:], i8[:], i8)', cache=True)
def lr_statistic(dur_1, dur_2, cens_1, cens_2, weightings):
    times = np.unique(np.hstack((dur_1, dur_2)))
    dur_1 = np.searchsorted(times, dur_1) + 1
    dur_2 = np.searchsorted(times, dur_2) + 1
    times_range = np.array([1, times.shape[0]], dtype=np.int32)

    bins = times_range[1] - times_range[0] + 1
    n_1_j = np.histogram(dur_1, bins=bins, range=times_range)[0]
    n_2_j = np.histogram(dur_2, bins=bins, range=times_range)[0]
    O_1_j = np.histogram(dur_1 * cens_1, bins=bins, range=times_range)[0]
    O_2_j = np.histogram(dur_2 * cens_2, bins=bins, range=times_range)[0]

    N_1_j = np.cumsum(n_1_j[::-1])[::-1]
    N_2_j = np.cumsum(n_2_j[::-1])[::-1]
    ind = np.where(N_1_j * N_2_j != 0)
    N_1_j = N_1_j[ind]
    N_2_j = N_2_j[ind]
    O_1_j = O_1_j[ind]
    O_2_j = O_2_j[ind]

    N_j = N_1_j + N_2_j
    O_j = O_1_j + O_2_j
    E_1_j = N_1_j * O_j / N_j
    res = np.zeros((N_j.shape[0], 3), dtype=np.float32)
    res[:, 1] = O_1_j - E_1_j
    res[:, 2] = E_1_j * (N_j - O_j) * N_2_j / (N_j * (N_j - 1))
    res[:, 0] = 1.0
    # if np.any(N_j <= 1):
    #     return 0.0
    if weightings == 2:
        res[:, 0] = N_j
    elif weightings == 3:
        res[:, 0] = np.sqrt(N_j)
    elif weightings == 4:
        res[:, 0] = np.cumprod((1.0 - O_j / (N_j + 1)))
    logrank = np.power((res[:, 0] * res[:, 1]).sum(), 2) / ((res[:, 0] * res[:, 0] * res[:, 2]).sum())
    return logrank


def weight_lr_fast(dur_A, dur_B, cens_A=None, cens_B=None, weightings=""):
    """
    Count weighted log-rank criteria

    Parameters
    ----------
    dur_A : array-like
        Time of occurred events from first sample.
    dur_B : array-like
        Time of occurred events from second sample.
    cens_A : array-like, optional
        Indicate of occurred events from first sample.
        The default is None (all events occurred).
    cens_B : array-like, optional
        Indicate of occurred events from second sample.
        The default is None (all events occurred).
    weightings : str, optional
        Weights of criteria. The default is "" (log-rank).
        Log-rank :math:'w = 1'
        Wilcoxon :math:'w = N_j'
        Tarone-ware :math:'w = \\sqrt(N_j)'
        Peto-peto :math:'w = \\fraq{1 - O_j}{N_j + 1}'

    Returns
    -------
    logrank : float
        Chi2 statistic value of weighted log-rank test

    """
    try:
        if cens_A is None:
            cens_A = np.ones(dur_A.shape[0])
        if cens_B is None:
            cens_B = np.ones(dur_B.shape[0])
        d = {"logrank": 1, "wilcoxon": 2, "tarone-ware": 3, "peto": 4}
        weightings = d.get(weightings, 1)
        logrank = lr_statistic(dur_A.astype("float64"),
                               dur_B.astype("float64"),
                               cens_A.astype("int64"),
                               cens_B.astype("int64"),
                               np.int64(weightings))
        return logrank
    except Exception as err:
        return 0.0



import pandas as pd
import numpy as np
import os
import copy
import tempfile

from sklearn.metrics import roc_auc_score

from .node import Node
from ..scheme import FilledSchemeStrategy
from .. import constants as cnt
import matplotlib.pyplot as plt

from graphviz import Digraph, set_jupyter_format
set_jupyter_format('png')


def format_to_pandas(X, columns):
    type_df = type(X)
    if type_df.__name__ == "DataFrame":
        return X[columns].reset_index(drop=True)
    elif type_df.__name__ == "ndarray":
        return pd.DataFrame(X, columns=columns)
    return None


def get_oversample(df, target=cnt.CENS_NAME):
    from imblearn.over_sampling import RandomOverSampler

    oversample = RandomOverSampler(sampling_strategy='minority', random_state=42)
    df_over, _ = oversample.fit_resample(df, df[target])
    return df_over


""" Functions of prunning """


def find_best_uncut(tree, X, y, target, mode_f, choose_f):
    span_leaf = tree.get_spanning_leaf_numbers()
    d = {}
    for el in span_leaf:
        y_pred = tree.predict(X, target=target, end_list=[el])
        d[el] = round(mode_f(y, y_pred), 4)

    new_leaf, val = choose_f(d.items(), key=lambda x: x[1])
    tree.delete_leaves_by_span([new_leaf])
    return tree, val


def cutted_tree(tree_, X, target, mode_f, choose_f, verbose=0):
    first_digits = lambda x: float(str(x)[:5])
    y = pd.to_numeric(X[target])
    tree = copy.deepcopy(tree_)
    best_metr = dict()
    best_tree = dict()
    y_pred = tree.predict(X, target=target)
    c = tree.get_leaf_numbers().shape[0]

    best_metr[c] = mode_f(y, y_pred)
    best_tree[c] = copy.deepcopy(tree)
    while len(tree.nodes) > 1:
        tree, val = find_best_uncut(tree, X, y, target, mode_f, choose_f)
        c = tree.get_leaf_numbers().shape[0]
        best_metr[c] = val
        best_tree[c] = copy.deepcopy(tree)

    best_metric = first_digits(choose_f(best_metr.values()))
    min_leaf = min([k for k, v in best_metr.items() if first_digits(v) == best_metric])

    if verbose > 0:
        plt.clf()
        plt.plot(list(best_metr.keys()), list(best_metr.values()), 'o')
        # plt.plot(list(best_metr.keys()), list(best_metr.values()), 'b')
        plt.xlabel("Количество листов")  # ("Leafs")
        plt.ylabel(f"Лучшее значение метрики {mode_f.__name__}")  # {target}")
        plt.title(f"Обрезка дерева по переменной {target}")
        plt.show()
        print(best_metr)
        print(best_metric, min_leaf)

    return best_tree[min_leaf]


class CRAID(object):
    """
    Survival decision tree model.

    Attributes
    ----------
    nodes : dict
        Dictionary of all tree's nodes (numbers from hierarchy)
    cut : boolean
        Flag of pruning
    balance : boolean
        Flag of source data balancing
    depth : int
        Maximal depth of nodes
    features : list
        Available features
    categ : list
        Names of categorical features
    random_state : int
        Fixed seed for building reproducibility
    name : str
        Model's name
    bins : array-like
        Points of timeline.
    info : dict
        Parameters for building nodes

    Methods
    -------

    fit : build decision tree with X, y data (iterative splitting node)
    predict : return values of features, rules or schemes
    predict_at_times : return survival or hazard function
    predict_schemes : return FilledSchemeStrategy or Scheme
    cut_tree : pruning function

    visualize : build graphviz Digraph for each node
    translate : Replace rules and features by dictionary

    get_leaf_numbers : return leaf numbers from nodes
    get_spanning_leaf_numbers : return pre-leaves numbers from nodes
    delete_leaves_by_span : set up pre-leaves from lists to leaves

    """
    def __init__(self, depth=0,
                 random_state=123,
                 features=[],
                 categ=[],
                 cut=False,
                 balance=None,
                 **info):
        self.info = info
        self.cut = cut
        self.balance = balance
        self.nodes = dict()
        self.depth = depth
        self.features = features
        self.categ = categ
        self.random_state = random_state
        self.name = f"CRAID_{self.random_state}"
        self.bins = []

    def update_params(self, X_tr):
        if not ("min_samples_leaf" in self.info):
            self.info["min_samples_leaf"] = 0.01
        if isinstance(self.info["min_samples_leaf"], float):
            self.info["min_samples_leaf"] = max(int(self.info["min_samples_leaf"] * X_tr.shape[0]), 1)

        cnt.set_seed(self.random_state)

        if self.balance in ["balance", "balance+correct"]:
            freq = X_tr[cnt.CENS_NAME].value_counts()
            self.correct_proba = freq[1] / (freq[1] + freq[0])  # or freq[1] / (freq[0])

            X_tr = get_oversample(X_tr, target=cnt.CENS_NAME)
        elif self.balance in ["balance+weights"]:
            freq = X_tr[cnt.CENS_NAME].value_counts()

            X_tr["weights_obs"] = np.where(X_tr[cnt.CENS_NAME], freq[0] / freq[1], 1)
            self.info["weights_feature"] = "weights_obs"
        elif self.balance in ["too_late"]:
            X_tr[cnt.CENS_NAME][X_tr[cnt.TIME_NAME] > np.quantile(X_tr[cnt.TIME_NAME], 0.7)] = False

        elif self.balance in ["only_log_rank"]:
            self.info["balance"] = True

        if "l_reg" in self.info:
            self.info["apr_time"] = X_tr[cnt.TIME_NAME]
            self.info["apr_event"] = X_tr[cnt.CENS_NAME]

    def fit(self, X, y):
        if len(self.features) == 0:
            self.features = X.columns
        self.bins = cnt.get_bins(time=y[cnt.TIME_NAME])  # , cens = y[cnt.CENS_NAME])
        X = X.reset_index(drop=True)
        X_tr = X.copy()
        X_tr[cnt.CENS_NAME] = y[cnt.CENS_NAME].astype(np.int32)
        X_tr[cnt.TIME_NAME] = y[cnt.TIME_NAME].astype(np.float32)

        self.update_params(X_tr)

        if self.cut:
            X_val = X_tr.sample(n=int(0.2 * X_tr.shape[0]), random_state=self.random_state)
            X_tr = X_tr.loc[X_tr.index.difference(X_val.index), :]

        self.nodes[0] = Node(X_tr, features=self.features, categ=self.categ, **self.info)
        stack_nodes = np.array([0], dtype=int)
        while stack_nodes.shape[0] > 0:
            node = self.nodes[stack_nodes[0]]
            stack_nodes = stack_nodes[1:]
            if node.depth >= self.depth:
                node.set_leaf()
                continue
            sub_nodes = node.split()
            if sub_nodes.shape[0] > 0:
                sub_numbers = np.array([len(self.nodes) + i for i in range(sub_nodes.shape[0])])
                for i in range(sub_nodes.shape[0]):
                    sub_nodes[i].numb = sub_numbers[i]
                self.nodes.update(dict(zip(sub_numbers, sub_nodes)))
                node.set_edges(sub_numbers)
                stack_nodes = np.append(stack_nodes, sub_numbers)

        if self.cut:
            self.cut_tree(X_val, cnt.CENS_NAME, mode_f=roc_auc_score, choose_f=max)

        return

    def predict(self, X, mode="target", target=cnt.TIME_NAME, end_list=[], bins=None):
        """
        Return values by mode & target

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events.
        mode : str, optional
            Mode of predicting. The default is "target".
            "surv" : return values of survival function in bins
            "hazard" : return values of hazard function in bins
            "target" : return values of feature (in target variable)
            "rules" : return full rules from node to leaf
        target : str or list, optional
            An aim of predicting. The default is occurred time.
        end_list : list, optional
            Numbers of endpoint nodes (for cutting)
        bins : array-like, optional
            Points of timeline

        Returns
        -------
        res : array-like
            Values by mode & target

        """
        X = format_to_pandas(X, self.features)
        num_node_to_key = dict(zip(sorted(self.nodes.keys()), range(len(self.nodes))))
        node_bin = np.zeros((X.shape[0], len(self.nodes)), dtype=bool)
        node_bin[:, 0] = 1
        shape = (X.shape[0])
        if not (bins is None):
            shape = (X.shape[0], len(bins))
        if target == "ch":
            shape = (X.shape[0], 3)
        if mode == "rules":
            res = np.full(shape, np.nan, dtype=object)
        else:
            res = np.empty(shape, dtype=object if target in self.categ else float)
            res[:] = np.nan
        for i in sorted(self.nodes.keys()):
            ind = node_bin[:, num_node_to_key[i]]
            if np.any(ind):
                X_loc = X[ind]
                if self.nodes[i].is_leaf or (i in end_list):
                    if target == "surv" or target == "hazard":
                        res[ind] = self.nodes[i].predict(X_loc, target, bins)
                    elif mode == "target":
                        res[ind] = self.nodes[i].predict(X_loc, target)
                    elif mode == "rules":
                        res[ind] = self.nodes[i].get_full_rule()
                else:
                    pred_edges = self.nodes[i].get_edges(X_loc)
                    for e in set(pred_edges):
                        node_bin[ind, num_node_to_key[e]] = pred_edges == e
        return res

    def predict_at_times(self, X, bins, mode="surv"):
        """
        Return survival or hazard function.

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events.
        bins : array-like
            Points of timeline.
        mode : str, optional
            Type of function. The default is "surv".
            "surv" : send building function in nodes
            "hazard" : send building function in nodes

        Returns
        -------
        res : array-like
            Vector of function values in times (bins)

        """
        X = format_to_pandas(X, self.features)
        pred = self.predict(X, target=mode, bins=bins)  # Reverse correction
        if self.balance == "balance+correct":
            if mode == "hazard":
                pred = pred * self.correct_proba
            elif mode == "surv":
                pred = pred ** self.correct_proba
        return pred

    def predict_schemes(self, X, scheme_feats):
        X = format_to_pandas(X, self.features)
        num_node_to_key = dict(zip(sorted(self.nodes.keys()), range(len(self.nodes))))
        node_bin = np.zeros((X.shape[0], len(self.nodes)), dtype=bool)
        node_bin[:, 0] = 1
        for i in sorted(self.nodes.keys()):
            i_num = num_node_to_key[i]
            ind = np.where(node_bin[:, i_num])[0]
            ind_x = X.index[ind]
            if ind.shape[0] > 0:
                if not self.nodes[i].is_leaf:
                    if self.nodes[i].rule_edges[0].get_feature() in scheme_feats:
                        for e in self.nodes[i].edges:
                            node_bin[ind, num_node_to_key[e]] = 1
                    else:
                        pred_edges = self.nodes[i].get_edges(X.loc[ind_x, :])
                        for e in set(pred_edges):
                            node_bin[ind, num_node_to_key[e]] = pred_edges == e
        leaf_keys = self.get_leaf_numbers()
        leaf_numb = np.array([num_node_to_key[leaf] for leaf in leaf_keys])

        ret_leaf_numbers = np.where(node_bin[:, leaf_numb], leaf_keys, np.inf)

        dict_leaf_scheme = {n_l: self.nodes[n_l].predict_scheme(None, scheme_feats)
                            for n_l in np.unique(ret_leaf_numbers) if n_l != np.inf}
        dict_str_fill = {}
        for leaf_list in np.unique(ret_leaf_numbers, axis=0):
            end_leaf = leaf_list[leaf_list != np.inf]
            sch_list = np.vectorize(dict_leaf_scheme.get)(end_leaf)
            dict_str_fill[str(end_leaf)] = FilledSchemeStrategy(sch_list)

        res = np.array(list(map(lambda leaf_list: dict_str_fill.get(str(leaf_list[leaf_list != np.inf]), np.nan),
                                ret_leaf_numbers)), dtype=object)
        return res

    def cut_tree(self, X, target, mode_f=roc_auc_score, choose_f=max):
        """
        Method of pruning tree.
        Find the best subtree that achieves the best value of the "mode_f" metric".

        Parameters
        ----------
        X : Pandas dataframe
            Contain input features of events.
        target : str
            Feature name for metric counting.
        mode_f : function, optional
            Metric for selecting. The default is roc_auc_score.
        choose_f : function, optional
            Type of best value (max or min). The default is max.

        """
        self.nodes = cutted_tree(self, X, target, mode_f, choose_f).nodes

    def visualize(self, path_dir=None, **kwargs):
        """
        Build graphviz representation of the tree

        Parameters
        ----------
        path_dir : str
        kwargs : dict
            Keyword arguments for nodes

        Returns
        -------
        dot : graphviz.Digraph
        """
        if path_dir is None:
            path_dir = os.getcwd()
        kwargs["bins"] = self.bins

        with tempfile.TemporaryDirectory() as tmp_dir:
            dot = Digraph(node_attr={'shape': 'none'})
            ordered_nodes = sorted(self.nodes.keys())
            for i in ordered_nodes:
                dot = self.nodes[i].set_dot_node(dot, path_dir=tmp_dir, **kwargs)
            for i in ordered_nodes:
                dot = self.nodes[i].set_dot_edges(dot)
            dot.render(os.path.join(path_dir, self.name), cleanup=True, format="png")
        return dot

    def translate(self, describe):
        """ Rename features for each node by dictionary """
        self.features = [describe.get(f, f) for f in self.features]
        self.categ = [describe.get(c, c) for c in self.categ]
        for i in self.nodes.keys():
            self.nodes[i].translate(describe)

    """ Cutting methods """

    def get_leaf_numbers(self):
        return np.array([i for i in self.nodes.keys() if self.nodes[i].is_leaf])

    def get_spanning_leaf_numbers(self):
        """ Get pre-termination nodes (have two leaves in edges) """
        leaves = self.get_leaf_numbers()
        return np.array([i for i in self.nodes.keys()
                         if np.intersect1d(self.nodes[i].edges, leaves).shape[0] == 2])

    def delete_leaves_by_span(self, list_span_leaf):
        """ Set pre-termination nodes as leaves """
        for i in list_span_leaf:
            for e in self.nodes[i].edges:
                del self.nodes[e]
            self.nodes[i].set_leaf()

    def tolerance_find_best(self, *args, **kwargs):
        pass



import numpy as np

from .. import metrics as metr
from .boosting import BoostingCRAID

class IBSBoostingCRAID(BoostingCRAID):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IBSBoostingCRAID"

    def count_model_weights(self, model, X_sub, y_sub):
        if self.all_weight:
            X_sub = self.X_train
            y_sub = self.y_train

        pred_sf = model.predict_at_times(X_sub, bins=self.bins, mode="surv")

        wei = metr.ibs(self.y_train, y_sub, pred_sf, self.bins, axis=0)
        betta = np.mean(wei)
        return wei, betta

    def update_weight(self, index, wei_i):
        if self.all_weight:
            self.weights = self.weights + wei_i
        else:
            self.weights[index] = (self.weights[index] + wei_i)
        self.weights = (self.weights - self.weights.min())
        # self.weights = (self.weights - self.weights.min()) / (self.weights.max() - self.weights.min())

    def get_aggreg(self, x):
        if self.aggreg_func == 'median':
            return np.median(x, axis=0)
        if self.aggreg_func == 'moda':
            res = np.median(x > 0.5, axis=0)
            return res
        if self.aggreg_func == 'complex':
            # a = np.max(x, axis=0)
            b = np.min(x, axis=0)
            res = np.mean(x, axis=0)
            # res[a == 1] = 1
            res[res < 0.01] = 0
            return res
        elif self.aggreg_func == 'wei':
            inv_wei = 1 / np.array(self.bettas)
            wei = inv_wei / sum(inv_wei)
            return np.sum((x.T * wei).T, axis=0)
        return np.mean(x, axis=0)


class IBSProbBoostingCRAID(BoostingCRAID):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "IBSProbBoostingCRAID"

    def count_model_weights(self, model, X_sub, y_sub):
        if self.all_weight:
            X_sub = self.X_train
            y_sub = self.y_train

        pred_sf = model.predict_at_times(X_sub, bins=self.bins, mode="surv")
        m = metr.IBS_DICT.get(self.ens_metric_name.upper(), metr.ibs)

        ibs_vals = m(self.y_train, y_sub, pred_sf, self.bins, axis=0)

        ibs_scaled = (ibs_vals - np.mean(ibs_vals))  # first scheme
        # ibs_scaled = (ibs_vals - np.mean(ibs_vals)) / np.std(ibs_vals) # second scheme
        wei = 1 / (1 + np.exp(-ibs_scaled))
        betta = np.mean(wei)  # np.mean(ibs_vals)/np.std(ibs_vals)
        return wei, betta

    def update_weight(self, index, wei_i):
        if self.all_weight:
            self.weights = self.weights * wei_i
        else:
            self.weights[index] = (self.weights[index] * wei_i)
        # self.weights = self.weights / np.sum(self.weights)  # second scheme

    def get_aggreg(self, x):
        if self.aggreg_func == 'median':
            return np.median(x, axis=0)
        if self.aggreg_func == 'moda':
            res = np.median(x > 0.5, axis=0)
            return res
        if self.aggreg_func == 'complex':
            # a = np.max(x, axis=0)
            b = np.min(x, axis=0)
            res = np.mean(x, axis=0)
            # res[a == 1] = 1
            res[res < 0.01] = 0
            return res
        elif self.aggreg_func == 'wei':
            inv_wei = 1 / np.array(self.bettas)
            wei = inv_wei / sum(inv_wei)
            return np.sum((x.T * wei).T, axis=0)
        return np.mean(x, axis=0)



import numpy as np
import random
from .. import metrics as metr
from .. import criteria as scrit
from ..criteria import chi2_sf
# from scipy.stats import chi2

import matplotlib.pyplot as plt
import seaborn as sns
sns.set()


class Scheme(object):
    def __init__(self, rule, times, cens, feat_means):
        self.rule = rule
        self.times = times
        self.cens = cens
        self.feat_means = feat_means
        self.shape = times.shape[0]
        self.kmf = metr.get_survival_func(self.times, self.cens)

    def join(self, sch):
        self.rule = f"({self.rule})|({sch.rule})"
        self.times = np.hstack([self.times, sch.times])
        self.cens = np.hstack([self.cens, sch.cens])
        self.feat_means = (self.feat_means + sch.feat_means) / 2
        return self

    def predict_proba(self):
        return round(np.mean(self.cens), 5)

    def predict_time(self):
        return round(np.mean(self.times), 5)

    def copy(self):
        return Scheme(self.rule, self.times, self.cens, self.feat_means)

    def get_str_rules(self):
        return self.rule

    def get_description(self):
        descr = f"\nКОЛ-ВО:{self.shape} "
        descr += f"ВЕРОЯТНОСТЬ СОБЫТИЯ:{self.predict_proba()} "
        descr += f"ВРЕМЯ СОБЫТИЯ:{self.predict_time()} "
        descr += "\n" + self.get_str_rules()
        return descr

    def get_subschemes(self, min_size=5, top=3):
        ret = [self.copy()]
        return ret

    def plot_sf(self, ax, show_censors=False, ci_show=False):
        return self.kmf.plot_survival_function(ax=ax, censor_styles={"marker": 'o', "ms": 6},
                                               show_censors=show_censors, ci_show=ci_show)


class FilledSchemeStrategy(object):
    def __init__(self, schemes_list):
        self.schemes_dict = {sch.get_str_rules(): sch for sch in schemes_list}

    def join(self, fss):
        for k, v in fss.schemes_dict.items():
            if k in self.schemes_dict.keys():
                self.schemes_dict[k].join(v)
            else:
                self.schemes_dict[k] = v

    def join_nearest_leaves(self, sign_thres=0.05, diff_func=random.random):
        def delete_k_from_dict(d, del_k):
            d_ = dict()
            for k_ in d.keys():
                if not (del_k in k_.split("#")):
                    d_[k_] = d[k_]
            return d_

        base = self.schemes_dict
        diff_dict = dict()
        for i1, l1 in enumerate(base.keys()):
            for i2, l2 in enumerate(base.keys()):
                if i2 > i1:
                    diff_dict[l1 + '#' + l2] = scrit.logrank(base[l1].times, base[l2].times, base[l1].cens,
                                                             base[l2].cens)
        while len(base) > 1:
            max_pair_key, min_stat_val = min(diff_dict.items(), key=lambda x: x[1])
            max_p_val = chi2_sf(min_stat_val, df=1)
            print('Максимальное P-value:', max_p_val)
            if max_p_val < sign_thres:
                break
            f_l, s_l = max_pair_key.split('#')
            new_sch = base[f_l].copy().join(base[s_l])
            new_sch_name = new_sch.get_str_rules()
            for k in [f_l, s_l]:
                diff_dict = delete_k_from_dict(diff_dict, k)
                del base[k]
            for k in base.keys():
                diff_dict[new_sch_name + '#' + k] = scrit.logrank(new_sch.times, base[k].times, new_sch.cens,
                                                                  base[k].cens)
            base[new_sch_name] = new_sch
            print('Цепочки схем:', f_l, s_l)
            print('Заменяются на:', new_sch_name)
        self.schemes_dict = base

    def get_flatten_sch(self, sort_by=None):
        schemes = []
        for k, v in self.schemes_dict.items():
            schemes += v.get_subschemes()
        # descending order
        if sort_by == "size":
            schemes.sort(key=lambda x: x.shape, reverse=True)
        elif sort_by == "proba":
            schemes.sort(key=lambda x: x.predict_proba(), reverse=True)
        elif sort_by == "time":
            schemes.sort(key=lambda x: x.predict_time(), reverse=True)
        return schemes

    def predict_best_scheme(self, sort_by="proba"):
        return self.get_flatten_sch(sort_by=sort_by)[0]

    def visualize_schemes(self, output_dir=None, sort_by=None, show_censors=False, ci_show=False):
        fig, ax = plt.subplots(figsize=(7, 5))
        schemes = self.get_flatten_sch(sort_by=sort_by)
        ind = 0
        legends = []

        for sch in schemes:
            ind += 1
            descr = f"Схема №{ind}" + sch.get_description()
            ax = sch.plot_sf(ax, show_censors=show_censors, ci_show=ci_show)
            if show_censors and (0 in sch.cens):
                legends.append("Лечение завершено")
            legends.append(descr)
            if ci_show:
                legends.append("Доверительный интервал")

        if len(legends) > 0:
            ax.get_legend().remove()
            if not (output_dir is None):
                plt.savefig(output_dir + "scheme_figure.png")
            legend = ax.legend(legends, bbox_to_anchor=(0, -0.1), loc="upper left", ncol=1, prop={'size': 10})
            fig_leg = legend.figure
            fig_leg.canvas.draw()
            bbox = legend.get_window_extent().transformed(fig_leg.dpi_scale_trans.inverted())
            if not (output_dir is None):
                fig_leg.savefig(output_dir + "scheme_legend.png", dpi="figure", bbox_inches=bbox)
                plt.close()
            else:
                plt.show()
        else:
            raise ValueError("Не выделены значимые схемы лечения")



import os
import random
import numpy as np
from collections import Counter

RANDOM_STATE = 123
""" int: Fixed seed for model reproducibility """

CENS_NAME = 'cens'
TIME_NAME = 'time'
""" str: Fixed names for internal representation """

TRANSLATOR = {
    "time": {"en": "Time", "ru": "Время"},
    "sf": {"en": "Survival probability", "ru": "Вероятность выживания"},
    "hf": {"en": "Hazard rate", "ru": "Риск события"},
    "events": {"en": "events", "ru": "события"},
    "mean time": {"en": "mean time", "ru": "среднее время"},
    "size": {"en": "size", "ru": "размер"},
    "density": {"en": "density", "ru": "Плотность"},
}

def set_seed(seed_value):
    os.environ['PYTHONHASHSEED'] = str(seed_value)
    random.seed(seed_value)
    np.random.seed(seed_value)


def get_y(cens, time):
    """
    Internal representation of target variable

    Parameters
    ----------
    cens : array-like, shape = (n_events,)
        Censoring flags.
    time : array-like, shape = (n_events,)
        Time of occurred events.

    Returns
    -------
    y : structured array
        Output containing the binary event indicator as first field,
        and time of event or time of censoring as second field.

    """
    cens, time = np.array(cens), np.array(time) 
    y = np.empty(dtype=[(CENS_NAME, bool), 
                        (TIME_NAME, np.float64)], 
                 shape=cens.shape[0])
    y[CENS_NAME] = cens
    y[TIME_NAME] = time
    return y


def pd_to_xy(df):
    """
    Splitting pandas dataframe to feature space and target variables

    Parameters
    ----------
    df : pandas DataFrame
        Must contain columns CENS_NAME and TIME_NAME.

    Returns
    -------
    X : pandas DataFrame
        Feature space with remaining features.
    y : structured array
        y containing the binary event indicator as first field,
        and time of event or time of censoring as second field.

    """
    X = df[list(set(df.columns) - {CENS_NAME, TIME_NAME})]
    y = get_y(df[CENS_NAME], df[TIME_NAME])
    return X, y


def get_bins(time, cens=None, mode='a', num_bins=100):
    """
    Generate array of time points in timeline (from sample)

    Parameters
    ----------
    time : array-like
        Time of occurred events.
    cens : array-like, optional
        Censoring flags. The default is None (all events occurred).
    mode : str, optional
        Method of generation. The default is 'a'.
        'a' : all points (from min to max)
        'q' : quantile points (quantity is based on num_bins)
    num_bins : int, optional
        Quantity of required points. The default is 100.

    Returns
    -------
    bins : array
        Timeline

    """
    if not (cens is None):
        time = time[np.where(cens)]
    time = time.astype(np.int32)
    bins = np.array([])
    if mode == 'q':
        bins = np.quantile(time, np.arange(num_bins) / num_bins)
    elif mode == 'a':
        bins = np.arange(time.min(), time.max()+1)  # all bins
        # bins = np.unique(np.quantile(time, np.arange(2.5, 97.5) / 100))  # stable quantile bins
        # t_max = np.quantile(time, 0.95)  # NEW
        # bins = np.arange(time.min(), t_max)  # NEW
    return bins


def mode(a):
    # vals, cnts = np.unique(a, return_counts=True, equal_nan=False)
    # if vals.shape[0] == 0:
    #     return np.nan
    # modes, counts = vals[cnts.argmax()], cnts.max()
    # return modes
    return max(Counter(a).most_common(), key=lambda x: x[1] if x[0] == x[0] else 0)[0]



from .bootstrap import BootstrapCRAID, ParallelBootstrapCRAID
from .boosting import BoostingCRAID, SumBoostingCRAID
from .proboosting import ProbBoostingCRAID
from .ibsboosting import IBSBoostingCRAID, IBSProbBoostingCRAID
from .clever_boosting import IBSCleverBoostingCRAID



from setuptools import setup, find_packages
from os.path import join, dirname
from pkg_resources import parse_requirements as _parse_requirements

SELF_DATASETS = []  # ["datasets/data/ONK/*", "datasets/data/COVID/*"]

PACKAGE_DATA = {"survivors": ["datasets/data/*"] + SELF_DATASETS}


def parse_requirements(filename):
    with open(filename) as fin:
        parsed_requirements = _parse_requirements(fin)
        requirements = [str(ir) for ir in parsed_requirements]
    return requirements

# Installing
# python setup.py install

# Final library will be established by path:
# C:\ProgramData\Anaconda3\envs\survive\Lib\site-packages

# Uploading to PYPI
# Old method (https://github.com/shirakiya/pypirc3)
# 1. pip install pypirc3
# 2. pypirc3 create -u **** -p ****
# 3. python setup.py register sdist upload

# Twine method (https://github.com/pypa/twine)
# 1. pip install twine
# 2. python setup.py sdist (recently delete the dist folder)
# 3. (test) twine upload --skip-existing --repository-url https://test.pypi.org/legacy/ dist/* -u **** -p ****
# 4. (release) twine upload --skip-existing --repository-url https://upload.pypi.org/legacy/ dist/* -u **** -p ****

# New twine method
# 1. python setup.py sdist bdist_wheel
# 2. twine upload --repository survivors dist/* (see in .pypirc file)

install_requires = parse_requirements("requirements/requirements.txt")

setup(
    name='survivors',
    version='1.7.1',
    license='BSD 3-Clause License',
    author='Iulii Vasilev',
    author_email='iuliivasilev@gmail.com',
    packages=find_packages(),
    long_description=open(join(dirname(__file__), 'README.rst')).read(),
    include_package_data=False,
    package_data=PACKAGE_DATA,
    python_requires='>=3.10',
    install_requires=install_requires,  # requirements_list
    keywords=["survival analysis", "time-to-event", "event data", "machine learning"]
)



import matplotlib.pyplot as plt
import seaborn as sns

custom_params = {"axes.spines.right": False, 'grid.color': 'lightgray', 'axes.grid': True, "axes.spines.top": False}
sns.set_theme(style="ticks", rc=custom_params)


def plot_survival_function(sf, bins, title=""):
    """
    Classic survival function plotting

    Parameters
    ----------
    sf: iterable (with len(sf) == len(bins))
        Survival probabilities
    bins: iterable
        Times of probabilities
    title: str

    """
    plt.step(bins, sf)
    plt.xlabel('Time')
    plt.ylabel('Survival probability')
    plt.title(title)
    plt.show()



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

custom_params = {"axes.spines.right": False, 'grid.color': 'lightgray', 'axes.grid': True, "axes.spines.top": False}
sns.set_theme(style="ticks", rc=custom_params)


def plot_func_comparison(y_true, y_preds, labels):
    """
    Plot multiple functions by method

    Parameters
    ----------
    y_true: structured np.ndarray with shape=(n, 2)
    y_preds: list of np.ndarray with shape=(n, n_bins)
    labels: list of str

    """
    linestyle = ('solid' if y_true["cens"] else 'dashed')
    fig, ax = plt.subplots(figsize=(8, 5))

    for y_pred, label in zip(y_preds, labels):
        ax.plot(y_pred, label=label)
    ax.hlines(1.0, 0, y_true["time"], color='r', linestyle=linestyle)
    ax.hlines(0.0, y_true["time"], len(y_preds[0]), color='r', linestyle=linestyle)
    ax.vlines(y_true["time"], 0, 1, color='r',
              linestyle=linestyle, linewidth=2, label="Target S(t)")
    ax.legend()
    #     plt.title(f'Прогноз для терминального события с временем T={y_true["time"]}')
    #     plt.xlabel('Время')
    #     plt.ylabel('Вероятность выживания')
    plt.title(f'Prediction for {"terminal" if y_true["cens"] else "censured"} event with time={y_true["time"]}')
    plt.xlabel('Time')
    plt.ylabel('Survival probability')
    plt.show()


def plot_metric_comparison(y_true, y_preds, labels, bins, metric):
    """
    Comparison quality metric in time

    Parameters
    ----------
    y_true: structured np.ndarray with shape=(n, 2)
    y_preds: list of np.ndarray with shape=(n, n_bins)
    labels: list of str
    bins: np.ndarray with shape=(n_bins)
    metric: function of internal metric

    """
    fig, ax = plt.subplots()
    m_name = metric.__name__

    for y_pred, label in zip(y_preds, labels):
        m_by_time = metric(y_true[np.newaxis], y_true[np.newaxis], y_pred[np.newaxis], bins, axis=1)
        m_val = metric(y_true[np.newaxis], y_true[np.newaxis], y_pred[np.newaxis], bins, axis=-1)
        ax.plot(m_by_time, label=f"{label} ({m_name}={m_val:.3f})")
    ax.legend()
    plt.title(f'{m_name}(t) for {"terminal" if y_true["cens"] else "censured"} event with time={y_true["time"]}')
    #     plt.title(f'{m_name}(t) для терминального события с временем T={y_true["time"]}')
    if m_name.find("auprc") == -1:
        plt.xlabel('Time')
    else:
        plt.xlabel(r'$\varphi$')
        plt.xticks(np.linspace(0, 100, 5), labels=np.linspace(0, 1, 5))
    plt.ylabel(f'{m_name}(t) value')
    #     plt.ylabel(f'Значение {m_name}(t)')
    plt.show()



import numpy as np
from numba import njit

from ..criteria import chi2_sf
from ..external import KaplanMeier, NelsonAalen, KaplanMeierZeroAfter, FullProbKM
from ..metrics import ibs_WW, auprc
from ..constants import get_y
# from scipy.stats import chi2
# import diptest

""" Auxiliary functions """


@njit('f4(f4[:], f4[:], f4[:], f4[:], u4, f4[:])', cache=True)
def lr_hist_statistic(time_hist_1, time_hist_2, cens_hist_1, cens_hist_2,
                      weightings, obs_weights):
    N_1_j = np.cumsum(time_hist_1[::-1])[::-1]
    N_2_j = np.cumsum(time_hist_2[::-1])[::-1]
    ind = np.where((cens_hist_1 + cens_hist_2 != 0) & (N_1_j * N_2_j != 0))[0]
    if ind.shape[0] == 0:
        return 0.0

    N_1_j = N_1_j[ind]  # + 1
    N_2_j = N_2_j[ind]  # + 1
    O_1_j = cens_hist_1[ind]
    O_2_j = cens_hist_2[ind]

    N_j = N_1_j + N_2_j
    O_j = O_1_j + O_2_j
    E_1_j = N_1_j * O_j / N_j
    res = np.zeros((N_j.shape[0], 3), dtype=np.float32)
    res[:, 1] = O_1_j - E_1_j  # np.abs(O_1_j - E_1_j)
    res[:, 2] = E_1_j * (N_j - O_j) * N_2_j / (N_j * (N_j))  # N_j - 1

    res[:, 0] = 1.0
    if weightings == 2:
        res[:, 0] = N_j
    elif weightings == 3:
        res[:, 0] = np.sqrt(N_j)
    elif weightings == 4:
        res[:, 0] = np.cumprod((1.0 - O_j / (N_j + 1)))
    elif weightings == 5:
        res[:, 0] = obs_weights[ind]
    elif weightings == 6:
        res[:, 0] = O_j/N_j
    elif weightings == 7:
        res[:, 0] = np.cumprod((1.0 - O_j / (N_j + 1)))
    elif weightings == 8:
        res[:, 0] = N_j/(N_1_j * N_2_j)
        # res[:, 0] = np.cumprod((1.0 - O_1_j / (N_1_j + 1))) - np.cumprod((1.0 - O_2_j / (N_2_j + 1)))
    var = (res[:, 0] * res[:, 0] * res[:, 2]).sum()
    num = res[:, 0] * res[:, 1]
    stat_val = np.power(num.sum(), 2) / var  # ((res[:, 0] * res[:, 0] * res[:, 2]).sum())

    # print("ALL:", N_j[0], "(", N_1_j[0], N_2_j[0], ")",
    #       "OBS:", np.sum(O_j), "(", O_1_j, O_2_j, ")",  # ind,
    #       "WEI:", res[:, 0],
    #       # "OBS:", np.sum(O_j), "(", np.sum(cens_hist_1), np.sum(cens_hist_2), ")",
    #       "STAT:", stat_val, "(", num, var, ")")

    if weightings == 7:
        res[:, 0] = 1 - res[:, 0]
        stat_val2 = np.power((res[:, 0] * res[:, 1]).sum(), 2) / ((res[:, 0] * res[:, 0] * res[:, 2]).sum())
        stat_val = max(stat_val, stat_val2)
    return stat_val  # It must be square of value (without sqrt)


def weight_hist_stat(time_hist_1, time_hist_2, cens_hist_1=None, cens_hist_2=None, weights_hist=None, weightings=""):
    try:
        if cens_hist_1 is None:
            cens_hist_1 = time_hist_1
        if cens_hist_2 is None:
            cens_hist_2 = time_hist_2
        if weights_hist is None:
            weights_hist = np.ones(time_hist_1.shape[0], dtype=np.float32)
        d = {"logrank": 1, "wilcoxon": 2, "tarone-ware": 3, "peto": 4, "weights": 5}
        d.update({"diff": 6, "maxcombo": 7, "symm_peto": 8})
        weightings = d.get(weightings, 1)

        logrank = lr_hist_statistic(time_hist_1.astype("float32"),
                                    time_hist_2.astype("float32"),
                                    cens_hist_1.astype("float32"),
                                    cens_hist_2.astype("float32"),
                                    np.uint32(weightings),
                                    weights_hist.astype("float32")
                                    )
    except Exception as err:
        # print(err)
        # print(time_hist_1, time_hist_2, cens_hist_1, cens_hist_2, weights_hist)
        # print(time_hist_1.shape, time_hist_2.shape, cens_hist_1.shape, cens_hist_2.shape, weights_hist.shape)
        logrank = 0.0
    return logrank


# """ Optimal criter splitting with balancing """
# def optimal_criter_split_hist(left_time_hist, left_cens_hist,
#                               right_time_hist, right_cens_hist,
#                               na_time_hist, na_cens_hist, weights_hist, criterion, dis_coef):
#     none_to = 0
#     max_stat_val = 1.0
#
#     if int(dis_coef) > 1:
#         dis_coef = int(dis_coef)
#         left_time_hist = left_time_hist + (dis_coef - 1) * left_cens_hist
#         right_time_hist = right_time_hist + (dis_coef - 1) * right_cens_hist
#         na_time_hist = na_time_hist + (dis_coef - 1) * na_cens_hist
#
#         left_cens_hist = left_cens_hist * dis_coef
#         right_cens_hist = right_cens_hist * dis_coef
#         na_cens_hist = na_cens_hist * dis_coef
#     elif dis_coef != 0:
#         if int(1/dis_coef) > 1:
#             cens_dc = int(1/dis_coef)
#             left_time_hist = (cens_dc - 1)*left_time_hist + left_cens_hist
#             right_time_hist = (cens_dc - 1)*right_time_hist + right_cens_hist
#             na_time_hist = (cens_dc - 1)*na_time_hist + (dis_coef - 1) * na_cens_hist
#
#             left_cens_hist = left_cens_hist * cens_dc
#             right_cens_hist = right_cens_hist * cens_dc
#             na_cens_hist = na_cens_hist * cens_dc
#
#     if na_time_hist.shape[0] > 0:
#         a = weight_hist_stat(left_time_hist + na_time_hist, right_time_hist,
#                              left_cens_hist + na_cens_hist, right_cens_hist,
#                              weights_hist, weightings=criterion)
#         b = weight_hist_stat(left_time_hist, right_time_hist + na_time_hist,
#                              left_cens_hist, right_cens_hist + na_cens_hist,
#                              weights_hist, weightings=criterion)
#         # Nans move to a leaf with maximal statistical value
#         none_to = int(a < b)
#         max_stat_val = max(a, b)
#     #         print(a, b)
#     else:
#         max_stat_val = weight_hist_stat(left_time_hist, right_time_hist,
#                                         left_cens_hist, right_cens_hist,
#                                         weights_hist, weightings=criterion)
#     return (max_stat_val, none_to)


def optimal_criter_split_hist(left_time_hist, left_cens_hist,
                              right_time_hist, right_cens_hist,
                              na_time_hist, na_cens_hist, weights_hist, criterion, dis_coef):
    none_to = 0
    # n1 = np.sum(left_time_hist)
    # n2 = np.sum(right_time_hist)
    # cf = n1 / (n1 + n2)
    # cf = 0.5

    if na_time_hist.shape[0] > 0:
        a = weight_hist_stat(left_time_hist + na_time_hist, right_time_hist,
                             left_cens_hist + na_cens_hist, right_cens_hist,
                             weights_hist, weightings=criterion)
        b = weight_hist_stat(left_time_hist, right_time_hist + na_time_hist,
                             left_cens_hist, right_cens_hist + na_cens_hist,
                             weights_hist, weightings=criterion)
        # Nans move to a leaf with maximal statistical value
        none_to = int(a < b)
        max_stat_val = max(a, b)
    else:
        max_stat_val = weight_hist_stat(left_time_hist, right_time_hist,
                                        left_cens_hist, right_cens_hist,
                                        weights_hist, weightings=criterion)
    return max_stat_val, none_to


def split_time_to_bins(time, event=None, apr_times=None, apr_events=None):
    # if apr_times is None:
    #     return np.searchsorted(np.unique(time), time)
    # return np.searchsorted(np.unique(apr_times), time)
    if apr_times is None:
        return np.searchsorted(np.arange(int(time.min() - 1), int(time.max() + 1)), time)
    return np.searchsorted(np.arange(int(apr_times.min() - 1), int(apr_times.max() + 1)), time)

    # if apr_times is None:
    #     n = np.clip(np.sum(event) // 5, 2, 20)
    #     return np.searchsorted(np.quantile(time[np.where(event)], np.arange(n + 1)/n), time)
    # n = np.clip(np.sum(apr_events) // 5, 2, 20)
    # return np.searchsorted(np.quantile(apr_times[np.where(apr_events)], np.arange(n + 1)/n), time)

#     if apr_times is None:
#         n = np.clip(time.shape[0] // 2, 2, 100)
#         return np.searchsorted(np.quantile(time, np.arange(n + 1)/n), time)
#     n = np.clip(apr_times.shape[0] // 2, 2, 100)
#     return np.searchsorted(np.quantile(apr_times, np.arange(n + 1)/n), time)


def get_attrs(max_stat_val, values, none_to, l_sh, r_sh, nan_sh, stat_diff):
    attrs = dict()
    attrs["stat_val"] = max_stat_val
    attrs["values"] = values
    attrs["stat_diff"] = stat_diff
    if none_to:
        attrs["pos_nan"] = [0, 1]
        attrs["min_split"] = min(l_sh, r_sh + nan_sh)
    else:
        attrs["pos_nan"] = [1, 0]
        attrs["min_split"] = min(l_sh + nan_sh, r_sh)
    return attrs


def transform_woe_np(x_feat, y):
    N_T = y.shape[0]
    N_D = y.sum()
    N_D_ = N_T - N_D
    x_uniq = np.unique(x_feat)
    x_dig = np.digitize(x_feat, x_uniq) - 1

    df_woe_iv = np.vstack([np.bincount(x_dig[y == 0], minlength=x_uniq.shape[0]),
                           np.bincount(x_dig[y == 1], minlength=x_uniq.shape[0])])
    all_0 = df_woe_iv[0].sum()
    all_1 = df_woe_iv[1].sum()

    p_bd = (df_woe_iv[1] + 1e-5) / (N_D + 1e-5)
    p_bd_ = (df_woe_iv[0] + 1e-5) / (N_D_ + 1e-5)
    p_b_d = (all_1 - df_woe_iv[1] + 1e-5) / (N_D + 1e-5)
    p_b_d_ = (all_0 - df_woe_iv[0] + 1e-5) / (N_D_ + 1e-5)

    woe_pl = np.log(p_bd / p_bd_)
    woe_mn = np.log(p_b_d / p_b_d_)
    descr_np = np.vstack([x_uniq, woe_pl - woe_mn])
    features_woe = dict(zip(descr_np[0], descr_np[1]))
    woe_x_feat = np.vectorize(features_woe.get)(x_feat)
    # calculate information value
    # iv = ((p_bd - p_bd_)*woe_pl).sum()
    return woe_x_feat, descr_np


def get_sa_hists(time, cens, minlength=1, weights=None):
    if time.shape[0] > 0:
        time_hist = np.bincount(time, minlength=minlength)
        cens_hist = np.bincount(time, weights=cens, minlength=minlength)
    else:
        time_hist, cens_hist = np.array([]), np.array([])
    return time_hist, cens_hist


def select_best_split_info(attr_dicts, type_attr, bonf=True, descr_woe=None, vals_categories=None):
    # attr_dicts = sorted(attr_dicts, key=lambda x: abs(x["stat_diff"]), reverse=True)[:max(1, len(attr_dicts)//2)]
    best_attr = max(attr_dicts, key=lambda x: x["stat_val"])

    best_attr["p_value"] = chi2_sf(best_attr["stat_val"], df=1)
    best_attr["sign_split"] = len(attr_dicts)
    if best_attr["sign_split"] > 0:
        if type_attr == "cont":
            best_attr["values"] = [f" <= {best_attr['values']}", f" > {best_attr['values']}"]
        # elif type_attr == "categ":
        #     best_attr["values"] = [f" in {e}" for e in best_attr["values"]]
        elif type_attr == "woe" or type_attr == "categ":
            ind = descr_woe[1] <= best_attr["values"]
            l, r = list(vals_categories[descr_woe[0, ind].astype(int)]), list(vals_categories[descr_woe[0, ~ind].astype(int)])
            # l, r = list(descr_woe[0, ind]), list(descr_woe[0, ~ind])
            best_attr["values"] = [f" in {e}" for e in [l, r]]
        if bonf:
            best_attr["p_value"] *= best_attr["sign_split"]
    return best_attr


def ranksums_hist(x, y):
    n1 = np.sum(x)
    n2 = np.sum(y)
    rank = np.arange(1, x.shape[0] + 1)
    s = np.dot(x, rank)
    expected = n1 * (n1+n2+1) / 2.0
    z = (s - expected) / np.sqrt(n1*n2*(n1+n2+1)/12.0)
    return z


def mw_hist(x, y):
    n1 = np.sum(x)
    n2 = np.sum(y)
    rank = np.arange(1, x.shape[0] + 1)
    R1 = np.dot(x, rank)
    U1 = R1 - n1*(n1+1)/2
    U2 = n1 * n2 - U1

    U = np.maximum(U1, U2)
    return U

# def diptest_hist(x, y):
#     res = 0
#     if np.sum(x) > 0:
#         t1 = np.repeat(np.arange(1, x.size + 1), x.astype(int))
#         res = max(res, diptest.dipstat(t1))
#     if np.sum(y) > 0:
#         t2 = np.repeat(np.arange(1, y.size + 1), y.astype(int))
#         res = max(res, diptest.dipstat(t2))
#     return res


def stdtest_hist(x, y):
    res = 0
    if np.sum(x) > 0:
        t1 = np.repeat(np.arange(1, x.size + 1), x.astype(int))
        res = max(res, t1.std())
    if np.sum(y) > 0:
        t2 = np.repeat(np.arange(1, y.size + 1), y.astype(int))
        res = max(res, t2.std())
    return res


def count_sf_diff(time, cens):
    c_time = np.cumsum(time[::-1])[::-1]
    sf = np.cumprod((1.0 - cens / (c_time + 1)))
    sf[c_time == 0] = 0.0
    return np.sum((sf - 0.5)**2) / sf.shape[0]  # max(1, bins[-1] - bins[0])


def cnttest_hist(time_1, cens_1, time_2, cens_2):
    d1 = count_sf_diff(time_1, cens_1)
    d2 = count_sf_diff(time_2, cens_2)
    return min(d1, d2)


def hist_best_attr_split(arr, criterion="logrank", type_attr="cont", weights=None, thres_cont_bin_max=100,
                         signif=1.0, signif_stat=0.0, min_samples_leaf=10, bonf=True, verbose=0, balance=False,
                         apr_time=None, apr_event=None, l_reg=0, **kwargs):
    """
    Find best split of sample by one feature (histogram realization)

    Parameters
    ----------
    arr: numpy ndarray
        Triplet of two target variables and one splitting feature
    criterion: str
        Weighting scheme in log-rank test
    type_attr: str
        Type of feature
        Mode :
            "cont" -- linear search in continuous values
            "categ" -- complete search of categorical variable combinations
            "woe" -- Weight Of Evidence mapping categorical to continuous
    weights: numpy ndarray
        Custom weights of observation significance
    thres_cont_bin_max: int
        Number of intermediate points
    signif: float
        Upper bound of the significance of the partition (p-value)
    signif_stat: float
        Lower bound of the significance of the partition (stat value)
    min_samples_leaf: int
        Minimum allowed number of observations in a leaf
    bonf: bool
        Using bonferroni correction
    verbose: int
        Enable verbose output
    balance: bool
        Enable balancing in log-rank
    apr_time: numpy ndarray
        Source times for log-rank regularization
    apr_event: numpy ndarray
        Source event indicator for log-rank regularization
    l_reg: float
        Regularization coefficient
    kwargs: dict
        Additional parameters
    Returns
    -------
    Dict with the best split of sample by feature
    """
    best_attr = {"stat_val": signif_stat, "p_value": signif,
                 "sign_split": 0, "values": [], "pos_nan": [1, 0]}
    if arr.shape[1] < 2 * min_samples_leaf:
        return best_attr
    # vals = arr[0].astype("float")
    # vals_categories = []
    if type_attr == "cont":
        vals = arr[0].astype("float")
        vals_categories = []
    else:
        arr_ = arr[0].astype(object)
        ind = arr_ == arr_
        vals = np.zeros_like(arr_, dtype=float)
        vals_categories, vals[ind] = np.unique(arr_[ind], return_inverse=True)
        vals[~ind] = np.nan
    cens = arr[1].astype("uint")
    dur = arr[2].astype("float")
    if np.sum(cens) == 0:
        # cens = np.ones_like(dur)
        return best_attr
    if weights is None:
        weights = np.ones(dur.shape, dtype=float)  # np.ones_like(dur)
        # kmf = KaplanMeierZeroAfter()
        # kmf.fit(dur, cens)
        # dd = np.unique(dur)
        # sf = kmf.survival_function_at_times(dd)
        # sf = np.repeat(sf[np.newaxis, :], dur.shape[0], axis=0)
        #
        # y = get_y(cens=cens, time=dur)
        # weights = ibs_WW(y, y, sf, dd, axis=0)
    weights_hist = None

    if apr_time is None:
        dur = split_time_to_bins(dur, cens)
        max_bin = dur.max()
        apr_t_distr = np.zeros(max_bin + 1)
        apr_e_distr = np.zeros(max_bin + 1)
    else:
        apr_time_1 = split_time_to_bins(apr_time, apr_event, dur, cens)  # , apr_time)
        dur = split_time_to_bins(dur, cens)
        max_bin = apr_time_1.max()
        apr_t_distr, apr_e_distr = get_sa_hists(apr_time_1, apr_event, minlength=max_bin + 1)
        apr_t_distr = apr_t_distr * l_reg * 0.5
        apr_e_distr = apr_e_distr * l_reg * 0.5

    # dur = split_time_to_bins(dur)
    # max_bin = dur.max()

    ind = np.isnan(vals)

    # split nan and not-nan
    dur_notna = dur[~ind]
    cens_notna = cens[~ind]
    vals_notna = vals[~ind]
    weights_notna = weights[~ind]

    dis_coef = 1
    if balance:
        dis_coef = (cens.shape[0] - np.sum(cens)) / np.sum(cens)
        # dis_coef = max(1, (cens.shape[0] - np.sum(cens)) // np.sum(cens))

    if dur_notna.shape[0] < min_samples_leaf:
        return best_attr

    descr_woe = None
    if type_attr == "woe" or type_attr == "categ":
        vals_notna, descr_woe = transform_woe_np(vals_notna, cens_notna)

    # find splitting values
    uniq_set = np.unique(vals_notna)
    if uniq_set.shape[0] > thres_cont_bin_max:
        uniq_set = np.quantile(vals_notna, [i / float(thres_cont_bin_max) for i in range(1, thres_cont_bin_max)])
    else:
        uniq_set = (uniq_set[:-1] + uniq_set[1:]) * 0.5
    uniq_set = np.unique(np.round(uniq_set, 3))

    index_vals_bin = np.digitize(vals_notna, uniq_set, right=True)

    # find global hist by times
    na_time_hist, na_cens_hist = get_sa_hists(dur[ind], cens[ind],
                                              minlength=max_bin + 1, weights=weights[ind])
    r_time_hist, r_cens_hist = get_sa_hists(dur_notna, cens_notna,
                                            minlength=max_bin + 1, weights=weights_notna)
    l_time_hist = np.zeros_like(r_time_hist, dtype=np.float32)  # + 1
    l_cens_hist = l_time_hist.copy()

    num_nan = ind.sum()
    num_r = dur_notna.shape[0]
    num_l = 0

    if criterion == "confident" or criterion == "confident_weights":
        kmf = KaplanMeier()
        if criterion == "confident_weights":
            kmf.fit(dur, cens, weights=weights)
        else:
            kmf.fit(dur, cens)
        ci = kmf.get_confidence_interval_()
        weights_hist = 1 / (ci[1:, 1] - ci[1:, 0] + 1e-5)  # (ci[1:, 1] + ci[1:, 0] + 1e-5)
        criterion = "weights"
    elif criterion == "fullprob":
        kmf = FullProbKM()
        kmf.fit(dur, cens)
        weights_hist = kmf.survival_function_at_times(np.unique(dur))
        criterion = "weights"
    elif criterion == "ibswei":
        kmf = KaplanMeierZeroAfter()
        dur_ = arr[2].copy()
        kmf.fit(dur_, cens)

        dd = np.unique(dur_)
        sf = kmf.survival_function_at_times(dd)
        sf = np.repeat(sf[np.newaxis, :], dd.shape[0], axis=0)

        y_ = get_y(cens=np.ones_like(dd), time=dd)
        y_["cens"] = True
        ibs_ev = ibs_WW(y_, y_, sf, dd, axis=0)
        y_["cens"] = False
        ibs_cn = ibs_WW(y_, y_, sf, dd, axis=0)

        ratio = np.sum(cens)/cens.shape[0]
        weights_hist = ibs_ev*ratio + ibs_cn*(1-ratio)
        criterion = "weights"
    elif criterion == "T-ET":
        kmf = KaplanMeierZeroAfter()
        dur_ = arr[2].copy()
        kmf.fit(dur_, cens)

        dd = np.unique(dur_)
        ET = np.trapz(kmf.survival_function_at_times(dd), dd)
        weights_hist = (dd - ET)  # **2
        criterion = "weights"
    elif criterion == "kde":
        na = NelsonAalen()
        na.fit(dur, cens, np.ones(len(dur)))
        weights_hist = na.get_smoothed_hazard_at_times(np.unique(dur))
        criterion = "weights"
    elif criterion == "wilcoxon-pred":
        if na_time_hist.shape[0] > 0:
            weights_hist = np.cumsum((r_time_hist + na_time_hist)[::-1])[::-1]
        else:
            weights_hist = np.cumsum(r_time_hist[::-1])[::-1]
        criterion = "weights"
    elif criterion == "tarone-ware-pred":
        if na_time_hist.shape[0] > 0:
            weights_hist = np.sqrt(np.cumsum((r_time_hist + na_time_hist)[::-1])[::-1])
        else:
            weights_hist = np.sqrt(np.cumsum(r_time_hist[::-1])[::-1])
        criterion = "weights"

    # elif weights is None:
    #     weights_hist = None
    # else:
    #     weights_hist = np.bincount(dur, weights=weights,  # /sum(weights),
    #                                minlength=max_bin + 1)
    #     weights_hist /= np.bincount(dur, minlength=max_bin + 1)  # np.sqrt()
        # weights_hist = np.cumsum(weights_hist[::-1])[::-1]  # np.sqrt()

        # weights_hist = weights_hist / weights_hist.sum()

    # for each split values get branches
    attr_dicts = []
    for u in np.unique(index_vals_bin)[:-1]:
        curr_mask = index_vals_bin == u
        curr_n = curr_mask.sum()
        curr_time_hist, curr_cens_hist = get_sa_hists(dur_notna[curr_mask], cens_notna[curr_mask],
                                                      minlength=max_bin + 1, weights=weights_notna[curr_mask])
        l_time_hist += curr_time_hist
        l_cens_hist += curr_cens_hist
        r_time_hist -= curr_time_hist
        r_cens_hist -= curr_cens_hist
        num_l += curr_n
        num_r -= curr_n

        if min(num_l, num_r) <= min_samples_leaf:
            continue

        max_stat_val, none_to = optimal_criter_split_hist(
            l_time_hist + apr_t_distr, l_cens_hist + apr_e_distr,
            r_time_hist + apr_t_distr, r_cens_hist + apr_e_distr,
            na_time_hist, na_cens_hist, weights_hist, criterion, dis_coef)

        if max_stat_val > signif_stat:
            stat_diff = 1
            # if na_time_hist.shape[0] > 0:
            #     stat_diff = stdtest_hist((l_time_hist + (1 - none_to) * na_time_hist),
            #                              (r_time_hist + none_to * na_time_hist))
            # else:
            #     stat_diff = stdtest_hist(l_time_hist,
            #                              r_time_hist)

            # if na_time_hist.shape[0] > 0:
            #     stat_diff = cnttest_hist((l_time_hist + (1 - none_to) * na_time_hist),
            #                              (l_cens_hist + (1 - none_to) * na_cens_hist),
            #                              (r_time_hist + none_to * na_time_hist),
            #                              (r_cens_hist + none_to * na_cens_hist))
            # else:
            #     stat_diff = cnttest_hist(l_time_hist, l_cens_hist, r_time_hist, r_cens_hist)

            attr_loc = get_attrs(max_stat_val, uniq_set[u], none_to, num_l, num_r, num_nan, stat_diff)
            attr_dicts.append(attr_loc)

    if len(attr_dicts) == 0:
        return best_attr

    # if na_time_hist.shape[0] > 0:
    #     sh_diff = count_sf_diff(r_time_hist + l_time_hist + na_time_hist,
    #                             r_cens_hist + l_cens_hist + na_cens_hist)
    # else:
    #     sh_diff = count_sf_diff(r_time_hist + l_time_hist,
    #                             r_cens_hist + l_cens_hist)
    #
    # attr_dicts = list(filter(lambda x: x["stat_diff"] > sh_diff / 2, attr_dicts))
    # if len(attr_dicts) == 0:
    #     return best_attr

    best_attr = select_best_split_info(attr_dicts, type_attr, bonf,
                                       descr_woe=descr_woe, vals_categories=vals_categories)
    if verbose > 0:
        print(best_attr["p_value"], len(uniq_set))
    return best_attr
